/* This file is part of the LWRPC library */
#include <lwrpc_internal.h>
#include <lwrpc.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

int is_finalized()
{
	int finalized = 0;

	MPI_Finalized(&finalized);
	
	return finalized;	
}

void enter_comm_cs()
{
	if (non_blocking) _lock_acquire(&commlock);
}

void leave_comm_cs()
{
	if (non_blocking) _lock_release(&commlock);
}

#define lwrpc_desc_size	sizeof(lwrpc_t)
MPI_Comm comm_out;

struct node_info *node_info;

static void call_broadcast(void *buffer, long count, MPI_Datatype datatype, long root);

void _lwrpc_comm_pre_init()
{
	static int already_called = -1;

	already_called++;
	if (already_called) {
		printf("_rpc_comm_pre_init has been called already\n");
		return;
	}

	node_info =(struct node_info *) calloc(1, MAX_NODES*sizeof(struct node_info));

	MPI_Barrier(comm_out);
}


void _lwrpc_comm_init()
{
	int i;
	int workers[64];
	int me = lwrpc_node_id();
	int nproc = lwrpc_num_nodes();
	
	workers[me] = kthreads;
	if (master_noworker) {
		if (me == 0) workers[me] = 0;
	}

	MPI_Allgather (&workers[me], 1, MPI_INT,
                                workers, 1, MPI_INT,
                                MPI_COMM_WORLD);

	MPI_Barrier(comm_out);
	
	for (i = 0; i < lwrpc_num_nodes(); i++) {
		node_info[i].nworkers = workers[i];
	}
	MPI_Barrier(comm_out);
}

static int _lwrpc_thread_id()
{
	if (pthread_equal(pthread_self(), server_thread))
		return MAX_CPUS;
	else
		return uth_get_vpid();
}

void send_arguments(int node, int tag, lwrpc_t *desc)
{
	int i;
	
	for (i = 0; i < desc->narg; i++)	{
		if ((desc->callway[i] == CALL_BY_VAL)||(desc->callway[i] == CALL_BY_VAD)||(desc->callway[i] == CALL_BY_COP)) {
			if (desc->quantity[i] == 1) continue;	
			enter_comm_cs();
			if (desc->homenode != desc->sourcenode)
				MPI_Send(&desc->temparg[i],desc->quantity[i], desc->dtype[i],node,tag,comm_out);
			else
				MPI_Send(&desc->localarg[i],desc->quantity[i], desc->dtype[i],node,tag,comm_out);
			leave_comm_cs();
		}
		else if ((desc->callway[i] == CALL_BY_REF)||(desc->callway[i] == CALL_BY_PTR)) {
			enter_comm_cs();
			if (desc->homenode != desc->sourcenode) {
				MPI_Send((void *)desc->temparg[i],desc->quantity[i],desc->dtype[i],node,tag,comm_out);
			} else {
				MPI_Send((void *)desc->localarg[i],desc->quantity[i],desc->dtype[i],node,tag,comm_out);
			}
			leave_comm_cs();
		}
		else 
			;
	}
}

void send_descriptor(int node, lwrpc_t *desc, int type)	
{
	int tag = _lwrpc_thread_id();
	int i;

	desc->sourcenode = lwrpc_node_id();
	desc->sourcevpid = tag;
	desc->type = type;

	enter_comm_cs();
	MPI_Send(desc, lwrpc_desc_size, MPI_CHAR,node,MAX_CPUS,comm_out);
	leave_comm_cs();
	switch (desc->type) {
		case LWRPC_ANSWER:
			for (i = 0; i < desc->narg; i++) {
				if ((desc->callway[i] == CALL_BY_REF) || (desc->callway[i] == CALL_BY_RES)) {
					enter_comm_cs();
					MPI_Send((void *)desc->temparg[i],desc->quantity[i],desc->dtype[i],desc->homenode,tag,comm_out);
					leave_comm_cs();
				}
				if (desc->quantity[i] > 1)
					free((void *)desc->temparg[i]);
			}
			return;
			break;
		default:
			if (desc->homenode == node) return;
			send_arguments(node, tag, desc);
			return;
			break;
	}
}

void direct_send_descriptor(int dummy, int sourcenode, int sourcevpid, lwrpc_t *desc)
{
	desc->sourcenode = lwrpc_node_id();
	desc->sourcevpid = MAX_CPUS+1;

	enter_comm_cs();
	MPI_Send(desc, lwrpc_desc_size, MPI_CHAR, sourcenode, sourcevpid, comm_out);
	leave_comm_cs();
	if (desc->homenode == sourcenode) return;

	send_arguments(sourcenode, sourcevpid, desc);
}

void receive_arguments(lwrpc_t *work, int tag)
{
	int i;
	int typesize;
	char *mem;
	MPI_Status status;
	int istat;
	
	for (i = 0; i < work->narg; i++) {
		if ((work->quantity[i] > 1)||((work->callway[i] != CALL_BY_VAL)&&(work->callway[i] != CALL_BY_VAD)&&(work->callway[i] != CALL_BY_COP))) {
			MPI_Type_size(work->dtype[i], &typesize);
			mem = (char *)malloc(work->quantity[i]*typesize);
			work->temparg[i] = mem;
			if ((work->callway[i] != CALL_BY_RES)&&(work->callway[i] != CALL_BY_RMA)) {
				enter_comm_cs();
				istat = MPI_Recv((void *)work->temparg[i], work->quantity[i], work->dtype[i], work->sourcenode, tag, comm_out, &status);
				leave_comm_cs();
			} 
			else if (work->callway[i] == CALL_BY_RES) {
				void memsetvalue(void *, double, int, MPI_Datatype);
				double setval;

				if ((work->operation[i] == 0)||(work->operation[i] == OP_SUM)) {
					setval = 0;
				}
				else if (work->operation[i] == OP_PROD) {
					setval = 1;
				}
				else {
					printf("should never reach here!\n");
					MPI_Abort(MPI_COMM_WORLD, 1);
				}
				memsetvalue((void *)work->temparg[i], setval, work->quantity[i], work->dtype[i]);
			}
			else {	
			}
		}	
		else {
			work->temparg[i] = work->localarg[i];
		}
	}

}

int receive_descriptor(int node, lwrpc_t *rpc)
{
	MPI_Status status;
	int istat;
	int tag = _lwrpc_thread_id();
	
	enter_comm_cs();
	istat = MPI_Recv(rpc, lwrpc_desc_size, MPI_CHAR, node, tag, comm_out, &status);
	leave_comm_cs();
	if (istat == MPI_SUCCESS)
		;
	else {
		rpc->type = LWRPC_NO_WORK;
		printf("MPI_Recv Failed!!!\n"); fflush(0);
		return 1;
	}

	if (rpc->type == LWRPC_NO_WORK) return 1;

	if (rpc->homenode == lwrpc_node_id()) {
	}
	else {
		receive_arguments(rpc, tag);
	}
	return 0;
}

static int _lwrpc_num_threads(int node_id)
{
	return node_info[node_id].nworkers;
}

int _lwrpc_total_num_threads()
{
	int i;
	int nodes = lwrpc_num_nodes();
	int sum_vp = 0;

	for (i = 0; i < nodes; i++) {
		sum_vp += _lwrpc_num_threads(i);
	}
	return sum_vp;
}

int global_thread_id_to_node_id(int global_thread_id)
{
	int i;
	int nodes = lwrpc_num_nodes();
	int sum_vp = 0;

	for (i = 0; i < nodes; i++) {
		sum_vp += _lwrpc_num_threads(i);
		if (global_thread_id < sum_vp) return i;
	}
	Error("target_to_node failed");
	return -1;
}

int local_thread_id_to_global_thread_id(int local_thread_id)
{
	int i;
	int mynode = lwrpc_node_id();
	int sum_vp = 0;

	for (i = 0; i < mynode; i++) {
		sum_vp += _lwrpc_num_threads(i);
	}
	return sum_vp + local_thread_id;
}


int global_thread_id_to_local_thread_id(int global_thread_id)
{
	int i;
	int mynode = global_thread_id_to_node_id(global_thread_id);
	int sum_vp = 0;

	for (i = 0; i < mynode; i++) {
		sum_vp += _lwrpc_num_threads(i);
	}
	return global_thread_id - sum_vp;
}

static void call_broadcast(void *buffer, long count, MPI_Datatype datatype, long root)
{
	enter_comm_cs();
	MPI_Bcast(buffer, count, datatype, root, comm_out);
	leave_comm_cs();
}

void lwrpc_broadcast(void *a, long count, MPI_Datatype datatype)
{
	long mynode = lwrpc_node_id();

	lwrpc_spmd(1, call_broadcast, 4,
			1, MPI_LONG, CALL_BY_VAL,
			1, MPI_LONG, CALL_BY_VAL,
			1, MPI_LONG, CALL_BY_VAL,
			1, MPI_LONG, CALL_BY_VAL,
			a, count, datatype, mynode);
}

void F77_FUNC_ (lwrpc_broadcast, LWRPC_BROADCAST) (void *a, long *count, MPI_Fint *datatype)
{
	MPI_Datatype dt;

	dt = MPI_Type_f2c(*datatype);
	lwrpc_broadcast(a, *count, dt);
}
