/* This file is part of the LWRPC library */
#include "uth.h"
#include "../include/utils.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define REUSE_THREADS

QUEUE_DEFINE(struct _uth_t, queue_t);

queue_t rsq;

void uth_to_rsq_end(struct _uth_t *uth)
{
	_enqueue_head(&rsq, uth);
}
	
struct _uth_t *uth_rsq_dequeue ()
{
	struct _uth_t *uth = NULL;

	_dequeue(&rsq, &uth);
	return uth;
}


/* Key for the thread-specific buffer */
static pthread_key_t _uth_key;
/* Once-only initialisation of the key */
static pthread_once_t _uth_key_once = PTHREAD_ONCE_INIT;

/* Free the thread-specific buffer */
static void _uth_destroy(void * buf)
{
        free(buf);
}

/* Allocate the key */
static void _uth_key_alloc(void)
{
        pthread_key_create(&_uth_key, _uth_destroy);
}

/* Allocate the thread-specific buffer */
void _uth_alloc(void)
{
        pthread_once(&_uth_key_once, _uth_key_alloc);
        pthread_setspecific(_uth_key, malloc(sizeof(void *)));
}


/* Return the thread-specific buffer */
void *_uth_get_self(void)
{
	void **ThisFib;

	ThisFib = (void **) pthread_getspecific(_uth_key);
	return (*ThisFib);
}


void _uth_set_self(void *uth)
{
	void **this_uth;

	this_uth = (void **) pthread_getspecific(_uth_key);
	*this_uth = uth;
}

/* Allocate the thread-specific buffer */
void _uth_init_self(void *uth)
{
	void **this_uth;

	_uth_alloc();
	this_uth = (void **) pthread_getspecific(_uth_key);
	*this_uth = uth;
}


uth_t uth_self()
{
	struct _uth_t *uth = (struct _uth_t *) _uth_get_self();

	/*printf("uth_self: %p\n", uth);*/
	return uth;
}

uth_t uth_self_ex(int vpid)
{
	return uth_self();
}

int uth_get_vpid()
{
	return uth_self()->vp_id;
}


void *uth_getarg(uth_t thread)
{
	return thread->arg;
}


void uth_setarg2(uth_t thread, void *arg)
{
        thread->arg = arg;
}


pthread_attr_t attr;

int uth_init(int stacksize)
{
	pthread_attr_init(&attr);

	pthread_attr_setstacksize(&attr, stacksize);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	
	_queue_init(&rsq);
	return 0;
}

int uth_vp_init(int vp, void *arg)
{
	struct _uth_t *uth = uth_create(0, arg);

	uth->pth_handler = pthread_self();
	uth->vp_id = vp;
	_uth_init_self(uth);	/* in thread local storage */

	return 0;
}


uth_t uth_create(void (*start)(void *arg), void *arg)
{
	struct _uth_t *uth = NULL;

#ifdef REUSE_THREADS
	uth = uth_rsq_dequeue();
#endif
	if (uth == NULL) {
		/*stksize = 8*1024*1024;*/
		uth = malloc(sizeof(struct _uth_t));	/* even recycle */
		pthread_cond_init(&uth->pth_cond, NULL);
		pthread_mutex_init (&uth->pth_mutex, NULL);
		uth->pth_handler = (pthread_t) NULL;
	}
	uth->func = start;
	uth->arg = arg;
	uth->next = NULL;

	return uth;
}


void uth_delete(uth_t thread)
{
	struct _uth_t *uth = thread;

#if 0
	pthread_cond_destroy(&uth->pth_cond);
	pthread_mutex_destroy (&uth->pth_mutex);
	uth->pth_handler = (pthread_t)NULL;
	free(uth);
#else
#ifndef REUSE_THREADS
	uth->pth_handler = (pthread_t)NULL;
#endif
	uth_to_rsq_end(uth);
#endif
}

void *_uth_main(void *arg)
{
	struct _uth_t *uth = (struct _uth_t *) arg;
	_uth_init_self(uth);

#ifdef REUSE_THREADS
	pthread_mutex_lock(&uth->pth_mutex);
	while (1) {
		pthread_mutex_unlock(&uth->pth_mutex);
		uth->func(uth->arg);
		/* if recycling is enabled, the thread is suspended here */
	}
	pthread_mutex_unlock(&uth->pth_mutex);
#else
	uth->func(uth->arg);
#endif

	return 0;
}

void uth_switchto_ex(int vp, struct _uth_t *old, struct _uth_t *new)
{
	new->vp_id = vp;

	if (old == NULL) {
#ifdef REUSE_THREADS
		struct _uth_t *old_uth = uth_self();
		pthread_mutex_lock(&old_uth->pth_mutex);	// here for better safety 
#endif

		pthread_mutex_lock(&new->pth_mutex);
		pthread_cond_signal(&new->pth_cond);	// unblock vp
		pthread_mutex_unlock(&new->pth_mutex);

		/* I am going to exit */
#ifdef REUSE_THREADS
		pthread_cond_wait(&old_uth->pth_cond, &old_uth->pth_mutex);
		pthread_mutex_unlock(&old_uth->pth_mutex);
#endif
	}
	else {
		pthread_mutex_lock(&old->pth_mutex);	// here for better safety 

		if (new->pth_handler ==  (pthread_t)NULL) {
			pthread_create(&new->pth_handler, &attr, _uth_main, new);

		}
		else {
			pthread_mutex_lock(&new->pth_mutex);
			pthread_cond_signal(&new->pth_cond);	// unblock vp
			pthread_mutex_unlock(&new->pth_mutex);
		}
		
		pthread_cond_wait(&old->pth_cond, &old->pth_mutex);
		pthread_mutex_unlock(&old->pth_mutex);
	}
}


#ifdef WIN32
int getpagesize()
{
	return 4096;
}
#endif

void uth_reinit(uth_t thread, void (*start)(void *arg), void *arg)
{
	thread->func = start;
	thread->arg = arg;
}


void uth_swap_thread_context(struct _uth_t **t1, struct _uth_t **t2)
{
	struct _uth_t *temp;

	temp = *t1;
	*t1 = *t2;
	*t2 = temp;
}


void uth_set_alloc_routine(void *func)
{
	/* not used */
}


#ifndef WIN32
#include <sys/time.h>
#include <time.h>  
#endif

double uth_gettime()
{
#ifdef WIN32
	return (double) GetTickCount()*0.001;
#else
	struct timeval t;
	gettimeofday(&t, NULL);
	return (double)t.tv_sec + (double)t.tv_usec*1.0E-6; 
#endif
}

